﻿// Copyright (C) Microsoft Corporation. All Rights Reserved.
// This code released under the terms of the Microsoft Public License
// (Ms-PL, http://opensource.org/licenses/ms-pl.html).

using System;
using System.Diagnostics.CodeAnalysis;
using System.Runtime.InteropServices;
using System.Text;

// Suppress warnings suggesting renaming the "Msi" class to "NativeMethods"
[assembly: SuppressMessage("Microsoft.Design", "CA1060:MovePInvokesToNativeMethodsClass", Scope = "member", Target = "Delay.Msi.#MsiCloseHandle(System.IntPtr)")]
[assembly: SuppressMessage("Microsoft.Design", "CA1060:MovePInvokesToNativeMethodsClass", Scope = "member", Target = "Delay.Msi.#MsiDoAction(System.IntPtr,System.String)")]
[assembly: SuppressMessage("Microsoft.Design", "CA1060:MovePInvokesToNativeMethodsClass", Scope = "member", Target = "Delay.Msi.#MsiFormatRecord(System.IntPtr,System.IntPtr,System.Text.StringBuilder,System.UInt32&)")]
[assembly: SuppressMessage("Microsoft.Design", "CA1060:MovePInvokesToNativeMethodsClass", Scope = "member", Target = "Delay.Msi.#MsiOpenPackage(System.String,System.IntPtr&)")]
[assembly: SuppressMessage("Microsoft.Design", "CA1060:MovePInvokesToNativeMethodsClass", Scope = "member", Target = "Delay.Msi.#MsiRecordGetFieldCount(System.IntPtr)")]
[assembly: SuppressMessage("Microsoft.Design", "CA1060:MovePInvokesToNativeMethodsClass", Scope = "member", Target = "Delay.Msi.#MsiRecordGetInteger(System.IntPtr,System.UInt32)")]
[assembly: SuppressMessage("Microsoft.Design", "CA1060:MovePInvokesToNativeMethodsClass", Scope = "member", Target = "Delay.Msi.#MsiRecordGetString(System.IntPtr,System.UInt32,System.Text.StringBuilder,System.UInt32&)")]
[assembly: SuppressMessage("Microsoft.Design", "CA1060:MovePInvokesToNativeMethodsClass", Scope = "member", Target = "Delay.Msi.#MsiSetExternalUI(Delay.Msi+InstallUiHandler,Delay.Msi+INSTALLLOGMODE,System.IntPtr)")]
[assembly: SuppressMessage("Microsoft.Design", "CA1060:MovePInvokesToNativeMethodsClass", Scope = "member", Target = "Delay.Msi.#MsiSetExternalUIRecord(Delay.Msi+InstallUiHandlerRecord,Delay.Msi+INSTALLLOGMODE,System.IntPtr,Delay.Msi+InstallUiHandlerRecord&)")]
[assembly: SuppressMessage("Microsoft.Design", "CA1060:MovePInvokesToNativeMethodsClass", Scope = "member", Target = "Delay.Msi.#MsiSetInternalUI(Delay.Msi+INSTALLUILEVEL,System.IntPtr&)")]
[assembly: SuppressMessage("Microsoft.Design", "CA1060:MovePInvokesToNativeMethodsClass", Scope = "member", Target = "Delay.Msi.#MsiSetProperty(System.IntPtr,System.String,System.String)")]

namespace Delay
{
    /// <summary>
    /// Class that exposes (parts of) the Windows Installer native API to managed applications.
    /// </summary>
    internal static class Msi
    {
        /// <summary>
        /// Opens a package to use with the functions that access the product database.
        /// </summary>
        /// <param name="szPackagePath">The path to the package.</param>
        /// <param name="hProduct">A pointer to a variable that receives the product handle.</param>
        /// <returns>ERROR value.</returns>
        [DllImport("Msi", CharSet = CharSet.Unicode)]
        public static extern ERROR MsiOpenPackage(string szPackagePath, out IntPtr hProduct);

        /// <summary>
        /// Closes an open installation handle.
        /// </summary>
        /// <param name="hAny">Specifies any open installation handle.</param>
        /// <returns>ERROR value.</returns>
        [DllImport("Msi")]
        public static extern ERROR MsiCloseHandle(IntPtr hAny);

        /// <summary>
        /// Executes a built-in action, custom action, or user-interface wizard action.
        /// </summary>
        /// <param name="hInstall">Handle to the installation provided to a DLL custom action or obtained through MsiOpenPackage, MsiOpenPackageEx, or MsiOpenProduct.</param>
        /// <param name="szAction">Specifies the action to execute.</param>
        /// <returns>ERROR value.</returns>
        [DllImport("Msi", CharSet = CharSet.Unicode)]
        public static extern ERROR MsiDoAction(IntPtr hInstall, string szAction);

        /// <summary>
        /// Sets the value for an installation property.
        /// </summary>
        /// <param name="hInstall">Handle to the installation provided to a DLL custom action or obtained through MsiOpenPackage, MsiOpenPackageEx, or MsiOpenProduct.</param>
        /// <param name="szName">Specifies the name of the property.</param>
        /// <param name="szValue">Specifies the value of the property.</param>
        /// <returns>ERROR value.</returns>
        [DllImport("Msi", CharSet = CharSet.Unicode)]
        public static extern ERROR MsiSetProperty(IntPtr hInstall, string szName, string szValue);

        /// <summary>
        /// Enables the installer's internal user interface.
        /// </summary>
        /// <param name="dwUILevel">Specifies the level of complexity of the user interface.</param>
        /// <param name="phWnd">Pointer to a window.</param>
        /// <returns>The previous user interface level is returned.</returns>
        [DllImport("Msi")]
        public static extern INSTALLUILEVEL MsiSetInternalUI(INSTALLUILEVEL dwUILevel, ref IntPtr phWnd);

        /// <summary>
        /// Enables an external user-interface handler.
        /// </summary>
        /// <param name="puiHandler">Specifies a callback function that conforms to the INSTALLUI_HANDLER specification.</param>
        /// <param name="dwMessageFilter">Specifies which messages to handle using the external message handler.</param>
        /// <param name="pvContext">Pointer to an application context that is passed to the callback function.</param>
        /// <returns>The return value is the previously set external handler, or zero (0) if there was no previously set handler.</returns>
        [DllImport("Msi")]
        public static extern InstallUiHandler MsiSetExternalUI(InstallUiHandler puiHandler, INSTALLLOGMODE dwMessageFilter, IntPtr pvContext);

        /// <summary>
        /// Defines a callback function that the installer calls for progress notification and error messages.
        /// </summary>
        /// <param name="pvContext">Pointer to an application context passed to the MsiSetExternalUI function.</param>
        /// <param name="iMessageType">Specifies a combination of one message box style, one message box icon type, one default button, and one installation message type.</param>
        /// <param name="szMessage">Specifies the message text.</param>
        /// <returns>-1 if an error was found in the message handler; 0 if no action was taken.</returns>
        public delegate int InstallUiHandler(IntPtr pvContext, uint iMessageType, string szMessage);

        /// <summary>
        /// Enables an external user-interface (UI) handler.
        /// </summary>
        /// <param name="puiHandler">Specifies a callback function that conforms to the INSTALLUI_HANDLER_RECORD specification.</param>
        /// <param name="dwMessageFilter">Specifies which messages to handle using the external message handler.</param>
        /// <param name="pvContext">A pointer to an application context that is passed to the callback function.</param>
        /// <param name="ppuiPrevHandler">Returns the pointer to the previously set callback function that conforms to the INSTALLUI_HANDLER_RECORD specification, or NULL if no callback is previously set.</param>
        /// <returns>ERROR value.</returns>
        [DllImport("Msi")]
        public static extern ERROR MsiSetExternalUIRecord(InstallUiHandlerRecord puiHandler, INSTALLLOGMODE dwMessageFilter, IntPtr pvContext, out InstallUiHandlerRecord ppuiPrevHandler);

        /// <summary>
        /// Defines a callback function that the installer calls for progress notification and error messages.
        /// </summary>
        /// <param name="pvContext">Pointer to an application context passed to the MsiSetExternalUIRecord function.</param>
        /// <param name="iMessageType">Specifies a combination of one message box style, one message box icon type, one default button, and one installation message type.</param>
        /// <param name="hRecord">Specifies a handle to the record object.</param>
        /// <returns>-1 if an error was found in the message handler; 0 if no action was taken.</returns>
        public delegate int InstallUiHandlerRecord(IntPtr pvContext, uint iMessageType, IntPtr hRecord);

        /// <summary>
        /// Formats record field data and properties using a format string.
        /// </summary>
        /// <param name="hInstall">Handle to the installation.</param>
        /// <param name="hRecord">Handle to the record to format.</param>
        /// <param name="szResultBuf">Pointer to the buffer that receives the null terminated formatted string.</param>
        /// <param name="pcchResultBuf">Pointer to the variable that specifies the size, in characters, of the buffer pointed to by the variable szResultBuf.</param>
        /// <returns>ERROR value.</returns>
        [DllImport("Msi", CharSet = CharSet.Unicode)]
        public static extern ERROR MsiFormatRecord(IntPtr hInstall, IntPtr hRecord, StringBuilder szResultBuf, ref uint pcchResultBuf);

        /// <summary>
        /// Returns the string value of a record field.
        /// </summary>
        /// <param name="hRecord">Handle to the record.</param>
        /// <param name="iField">Specifies the field requested.</param>
        /// <param name="szValueBuf">Pointer to the buffer that receives the null terminated string containing the value of the record field.</param>
        /// <param name="pcchValueBuf">Pointer to the variable that specifies the size, in characters, of the buffer pointed to by the variable szValueBuf.</param>
        /// <returns>ERROR value.</returns>
        [DllImport("Msi", CharSet = CharSet.Unicode)]
        public static extern ERROR MsiRecordGetString(IntPtr hRecord, uint iField, StringBuilder szValueBuf, ref uint pcchValueBuf);

        /// <summary>
        /// Returns the integer value from a record field.
        /// </summary>
        /// <param name="hRecord">Handle to a record.</param>
        /// <param name="iField">Specifies the field of the record from which to obtain the value.</param>
        /// <returns>If the function succeeds, the return value is the integer value of the field. The MsiRecordGetInteger function returns MSI_NULL_INTEGER if the field is null or if the field is a string that cannot be converted to an integer.</returns>
        [DllImport("Msi")]
        public static extern int MsiRecordGetInteger(IntPtr hRecord, uint iField);

        /// <summary>
        /// Returns the number of fields in a record.
        /// </summary>
        /// <param name="hRecord">Handle to a record.</param>
        /// <returns>If the function succeeds, the return value is the number of fields in the record. If the function is given an invalid or inactive handle, it returns -1 or 0xFFFFFFFF.</returns>
        [DllImport("Msi")]
        public static extern uint MsiRecordGetFieldCount(IntPtr hRecord);

        /// <summary>
        /// Integer value reserved for null.
        /// </summary>
        /// <remarks>Used by MsiRecordGetInteger.</remarks>
        public const int MSI_NULL_INTEGER = int.MinValue;

        /// <summary>
        /// Specifies the level of complexity of the user interface.
        /// </summary>
        public enum INSTALLUILEVEL
        {
            /// <summary>
            /// No change in the UI level. However, if phWnd is not Null, the parent window can change.
            /// </summary>
            NOCHANGE = 0,
            /// <summary>
            /// The installer chooses an appropriate user interface level.
            /// </summary>
            DEFAULT = 1,
            /// <summary>
            /// Completely silent installation.
            /// </summary>
            NONE = 2,
            /// <summary>
            /// Simple progress and error handling.
            /// </summary>
            BASIC = 3,
            /// <summary>
            /// Authored user interface with wizard dialog boxes suppressed.
            /// </summary>
            REDUCED = 4,
            /// <summary>
            /// Authored user interface with wizards, progress, and errors.
            /// </summary>
            FULL = 5,
            /// <summary>
            /// If combined with any above value, the installer displays a modal dialog box at the end of a successful installation or if there has been an error. No dialog box is displayed if the user cancels.
            /// </summary>
            ENDDIALOG = 0x80,
            /// <summary>
            /// If combined with the BASIC value, the installer shows simple progress dialog boxes but does not display any modal dialog boxes or error dialog boxes.
            /// </summary>
            PROGRESSONLY = 0x40,
            /// <summary>
            /// If combined with the BASIC value, the installer shows simple progress dialog boxes but does not display a Cancel button on the dialog. This prevents users from canceling the install.
            /// </summary>
            HIDECANCEL = 0x20,
            /// <summary>
            /// If this value is combined with the NONE value, the installer displays only the dialog boxes used for source resolution. No other dialog boxes are shown. This value has no effect if the UI level is not NONE. It is used with an external user interface designed to handle all of the UI except for source resolution. In this case, the installer handles source resolution.
            /// </summary>
            SOURCERESONLY = 0x100,
            /// <summary>
            /// Show UAC prompt even if quiet.
            /// </summary>
            UACONLY = 0x200,
        }

        /// <summary>
        /// Installation message type.
        /// </summary>
        public enum INSTALLMESSAGE
        {
            /// <summary>
            /// Premature termination.
            /// </summary>
            FATALEXIT = 0x00000000,
            /// <summary>
            /// Formatted error message.
            /// </summary>
            ERROR = 0x01000000,
            /// <summary>
            /// Formatted warning message.
            /// </summary>
            WARNING = 0x02000000,
            /// <summary>
            /// User request message.
            /// </summary>
            USER = 0x03000000,
            /// <summary>
            /// Informative message for log.
            /// </summary>
            INFO = 0x04000000,
            /// <summary>
            /// List of files currently in use that must be closed before being replaced.
            /// </summary>
            FILESINUSE = 0x05000000,
            /// <summary>
            /// Request to determine a valid source location.
            /// </summary>
            RESOLVESOURCE = 0x06000000,
            /// <summary>
            /// Insufficient disk space message.
            /// </summary>
            OUTOFDISKSPACE = 0x07000000,
            /// <summary>
            /// Start of action message. This message includes the action name and description.
            /// </summary>
            ACTIONSTART = 0x08000000,
            /// <summary>
            /// Formatted data associated with the individual action item.
            /// </summary>
            ACTIONDATA = 0x09000000,
            /// <summary>
            /// Progress gauge information. This message includes information on units so far and total number of units.
            /// </summary>
            PROGRESS = 0x0A000000,
            /// <summary>
            /// Formatted dialog information for user interface.
            /// </summary>
            COMMONDATA = 0x0B000000,
            /// <summary>
            /// Sent prior to UI initialization, no string data.
            /// </summary>
            INITIALIZE = 0x0C000000,
            /// <summary>
            /// Sent after UI termination, no string data.
            /// </summary>
            TERMINATE = 0x0D000000,
            /// <summary>
            /// Sent prior to display of authored dialog or wizard.
            /// </summary>
            SHOWDIALOG = 0x0E000000,
            /// <summary>
            /// List of files currently in use that must be closed before being replaced. Available beginning with Windows Installer version 4.0. For more information about this message see Using Restart Manager with an External UI.
            /// </summary>
            RMFILESINUSE = 0x19000000,
            /// <summary>
            /// Sent prior to installation of product.
            /// </summary>
            INSTALLSTART = 0x1A000000,
            /// <summary>
            /// Sent after installation of product.
            /// </summary>
            INSTALLEND = 0x1B000000,
        }

        /// <summary>
        /// Mask of all valid bits for an INSTALLMESSAGE value.
        /// </summary>
        public const int INSTALLMESSAGE_MASK = 0x1f000000;

        /// <summary>
        /// Specifies which messages to handle using the external message handler.
        /// </summary>
        [Flags]
        public enum INSTALLLOGMODE
        {
            /// <summary>
            /// Premature termination of installation.
            /// </summary>
            FATALEXIT = (1 << (INSTALLMESSAGE.FATALEXIT >> 24)),
            /// <summary>
            /// The error messages are logged.
            /// </summary>
            ERROR = (1 << (INSTALLMESSAGE.ERROR >> 24)),
            /// <summary>
            /// The warning messages are logged.
            /// </summary>
            WARNING = (1 << (INSTALLMESSAGE.WARNING >> 24)),
            /// <summary>
            /// The user requests are logged.
            /// </summary>
            USER = (1 << (INSTALLMESSAGE.USER >> 24)),
            /// <summary>
            /// The status messages that are not displayed are logged.
            /// </summary>
            INFO = (1 << (INSTALLMESSAGE.INFO >> 24)),
            /// <summary>
            /// Request to determine a valid source location.
            /// </summary>
            RESOLVESOURCE = (1 << (INSTALLMESSAGE.RESOLVESOURCE >> 24)),
            /// <summary>
            /// There is insufficient disk space.
            /// </summary>
            OUTOFDISKSPACE = (1 << (INSTALLMESSAGE.OUTOFDISKSPACE >> 24)),
            /// <summary>
            /// The start of new installation actions are logged.
            /// </summary>
            ACTIONSTART = (1 << (INSTALLMESSAGE.ACTIONSTART >> 24)),
            /// <summary>
            /// The data record with the installation action is logged.
            /// </summary>
            ACTIONDATA = (1 << (INSTALLMESSAGE.ACTIONDATA >> 24)),
            /// <summary>
            /// The parameters for user-interface initialization are logged.
            /// </summary>
            COMMONDATA = (1 << (INSTALLMESSAGE.COMMONDATA >> 24)),
            /// <summary>
            /// The Progress bar information. This message includes information about units so far and total number of units. This message is only sent to an external user interface and is not logged. For more information, see MsiProcessMessage.
            /// </summary>
            PROGRESS = (1 << (INSTALLMESSAGE.PROGRESS >> 24)),
            /// <summary>
            /// If this is not a quiet installation, then the basic UI is initialized. If this is a full UI installation, the Full UI is not yet initialized. This message is only sent to an external user interface and is not logged.
            /// </summary>
            INITIALIZE = (1 << (INSTALLMESSAGE.INITIALIZE >> 24)),
            /// <summary>
            /// If a full UI is being used, the full UI has ended. If this is not a quiet installation, the basic UI has not ended. This message is only sent to an external user interface and is not logged.
            /// </summary>
            TERMINATE = (1 << (INSTALLMESSAGE.TERMINATE >> 24)),
            /// <summary>
            /// Sent prior to display of the Full UI dialog. This message is only sent to an external user interface and is not logged.
            /// </summary>
            SHOWDIALOG = (1 << (INSTALLMESSAGE.SHOWDIALOG >> 24)),
            /// <summary>
            /// Files in use information. When this message is received, a FilesInUse Dialog should be displayed.
            /// </summary>
            FILESINUSE = (1 << (INSTALLMESSAGE.FILESINUSE >> 24)),
            /// <summary>
            /// Files in use information. When this message is received, a MsiRMFilesInUse Dialog should be displayed.
            /// </summary>
            RMFILESINUSE = (1 << (INSTALLMESSAGE.RMFILESINUSE >> 24)),
            /// <summary>
            /// Installation of product begins. The message contains the product's ProductName and ProductCode.
            /// </summary>
            INSTALLSTART = (1 << (INSTALLMESSAGE.INSTALLSTART >> 24)),
            /// <summary>
            /// Installation of product ends. The message contains the product's ProductName, ProductCode, and return value.
            /// </summary>
            INSTALLEND = (1 << (INSTALLMESSAGE.INSTALLEND >> 24)),
        }

        /// <summary>
        /// Error codes.
        /// </summary>
        public enum ERROR
        {
            /// <summary>
            /// The operation completed successfully.
            /// </summary>
            SUCCESS = 0,
            /// <summary>
            /// The system cannot find the file specified.
            /// </summary>
            FILE_NOT_FOUND = 2,
            /// <summary>
            /// The system cannot find the path specified.
            /// </summary>
            PATH_NOT_FOUND = 3,
            /// <summary>
            /// The handle is invalid.
            /// </summary>
            INVALID_HANDLE = 6,
            /// <summary>
            /// The data is invalid.
            /// </summary>
            INVALID_DATA = 13,
            /// <summary>
            /// The parameter is incorrect.
            /// </summary>
            INVALID_PARAMETER = 87,
            /// <summary>
            /// This function is not supported on this system.
            /// </summary>
            CALL_NOT_IMPLEMENTED = 120,
            /// <summary>
            /// More data is available.
            /// </summary>
            MORE_DATA = 234,
            /// <summary>
            /// User cancelled installation.
            /// </summary>
            INSTALL_USEREXIT = 1602,
            /// <summary>
            /// Fatal error during installation.
            /// </summary>
            INSTALL_FAILURE = 1603,
            /// <summary>
            /// Installation suspended, incomplete.
            /// </summary>
            INSTALL_SUSPEND = 1604,
            /// <summary>
            /// Handle is in an invalid state.
            /// </summary>
            INVALID_HANDLE_STATE = 1609,
            /// <summary>
            /// The configuration data for this product is corrupt. Contact your support personnel.
            /// </summary>
            BAD_CONFIGURATION = 1610,
            /// <summary>
            /// Another installation is already in progress. Complete that installation before proceeding with this install.
            /// </summary>
            INSTALL_ALREADY_RUNNING = 1618,
            /// <summary>
            /// This installation package could not be opened. Verify that the package exists and that you can access it, or contact the application vendor to verify that this is a valid Windows Installer package.
            /// </summary>
            INSTALL_PACKAGE_OPEN_FAILED = 1619,
            /// <summary>
            /// Function could not be executed.
            /// </summary>
            FUNCTION_NOT_CALLED = 1626,
            /// <summary>
            /// Function failed during execution.
            /// </summary>
            FUNCTION_FAILED = 1627,
            /// <summary>
            /// The Windows Installer does not permit installation from a Remote Desktop Connection.
            /// </summary>
            INSTALL_REMOTE_PROHIBITED = 1645,
        }
    }
}
